@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>
 
 import atexit
@@ -11,38 +10,44 @@ import tempfile
 import warnings
 from collections import defaultdict
 from contextlib import suppress
+from enum import Enum, auto
 from functools import partial
 from gettext import gettext as _
-from typing import DefaultDict, Dict, Iterable, List, Optional, Tuple, Union
+from typing import (
+    Any, DefaultDict, Dict, Iterable, Iterator, List, Optional, Tuple, Union
+)
 
 from kitty.cli import CONFIG_HELP, parse_args
 from kitty.cli_stub import DiffCLIOptions
-from kitty.conf.utils import KittensKeyAction
+from kitty.conf.utils import KeyAction
 from kitty.constants import appname
 from kitty.fast_data_types import wcswidth
-from kitty.key_encoding import RELEASE, KeyEvent, enter_key, key_defs as K
-from kitty.options_stub import DiffOptions
+from kitty.key_encoding import EventType, KeyEvent
 from kitty.utils import ScreenSize
 
-from . import global_data
-from .collect import (
-    Collection, create_collection, data_for_path, lines_for_path, sanitize,
-    set_highlight_data
-)
-from .config import init_config
-from .patch import Differ, Patch, set_diff_command, worker_processes
-from .render import (
-    ImagePlacement, ImageSupportWarning, Line, LineRef, Reference, render_diff
-)
-from .search import BadRegex, Search
 from ..tui.handler import Handler
 from ..tui.images import ImageManager, Placement
 from ..tui.line_edit import LineEdit
 from ..tui.loop import Loop
 from ..tui.operations import styled
+from . import global_data
+from .collect import (
+    Collection, add_remote_dir, create_collection, data_for_path,
+    lines_for_path, sanitize, set_highlight_data
+)
+from .config import init_config
+from .options.types import Options as DiffOptions
+from .patch import Differ, Patch, set_diff_command, worker_processes
+from .render import (
+    ImagePlacement, ImageSupportWarning, Line, LineRef, Reference, render_diff
+)
+from .search import BadRegex, Search
 
 try:
-    from .highlight import initialize_highlighter, highlight_collection, DiffHighlight
+    from .highlight import (
+        DiffHighlight, get_highlight_processes, highlight_collection,
+        initialize_highlighter
+    )
     has_highlighter = True
     DiffHighlight
 except ImportError:
@@ -51,9 +56,24 @@ except ImportError:
     def highlight_collection(collection: 'Collection', aliases: Optional[Dict[str, str]] = None) -> Union[str, Dict[str, 'DiffHighlight']]:
         return ''
 
+    def get_highlight_processes() -> Iterator[int]:
+        if has_highlighter:
+            yield -1
 
-INITIALIZING, COLLECTED, DIFFED, COMMAND, MESSAGE = range(5)
-ESCAPE = K['ESCAPE']
+
+class State(Enum):
+    initializing = auto()
+    collected = auto()
+    diffed = auto()
+    command = auto()
+    message = auto()
+
+
+class BackgroundWork(Enum):
+    none = auto()
+    collecting = auto()
+    diffing = auto()
+    highlighting = auto()
 
 
 def generate_diff(collection: Collection, context: int) -> Union[str, Dict[str, Patch]]:
@@ -74,7 +94,7 @@ class DiffHandler(Handler):
     image_manager_class = ImageManager
 
     def __init__(self, args: DiffCLIOptions, opts: DiffOptions, left: str, right: str) -> None:
-        self.state = INITIALIZING
+        self.state = State.initializing
         self.message = ''
         self.current_search_is_regex = True
         self.current_search: Optional[Search] = None
@@ -88,18 +108,22 @@ class DiffHandler(Handler):
         if self.current_context_count < 0:
             self.current_context_count = self.original_context_count = self.opts.num_context_lines
         self.highlighting_done = False
+        self.doing_background_work = BackgroundWork.none
         self.restore_position: Optional[Reference] = None
         for key_def, action in self.opts.key_definitions.items():
-            self.add_shortcut(action, *key_def)
+            self.add_shortcut(action, key_def)
 
-    def perform_action(self, action: KittensKeyAction) -> None:
+    def terminate(self, return_code: int = 0) -> None:
+        self.quit_loop(return_code)
+
+    def perform_action(self, action: KeyAction) -> None:
         func, args = action
         if func == 'quit':
-            self.quit_loop(0)
+            self.terminate()
             return
-        if self.state <= DIFFED:
+        if self.state.value <= State.diffed.value:
             if func == 'scroll_by':
-                return self.scroll_lines(int(args[0]))
+                return self.scroll_lines(int(args[0] or 0))
             if func == 'scroll_to':
                 where = str(args[0])
                 if 'change' in where:
@@ -119,7 +143,7 @@ class DiffHandler(Handler):
                 elif to == 'default':
                     new_ctx = self.original_context_count
                 else:
-                    new_ctx += int(to)
+                    new_ctx += int(to or 0)
                 return self.change_context_count(new_ctx)
             if func == 'start_search':
                 self.start_search(bool(args[0]), bool(args[1]))
@@ -128,8 +152,9 @@ class DiffHandler(Handler):
     def create_collection(self) -> None:
 
         def collect_done(collection: Collection) -> None:
+            self.doing_background_work = BackgroundWork.none
             self.collection = collection
-            self.state = COLLECTED
+            self.state = State.collected
             self.generate_diff()
 
         def collect(left: str, right: str) -> None:
@@ -137,15 +162,17 @@ class DiffHandler(Handler):
             self.asyncio_loop.call_soon_threadsafe(collect_done, collection)
 
         self.asyncio_loop.run_in_executor(None, collect, self.left, self.right)
+        self.doing_background_work = BackgroundWork.collecting
 
     def generate_diff(self) -> None:
 
         def diff_done(diff_map: Union[str, Dict[str, Patch]]) -> None:
+            self.doing_background_work = BackgroundWork.none
             if isinstance(diff_map, str):
                 self.report_traceback_on_exit = diff_map
-                self.quit_loop(1)
+                self.terminate(1)
                 return
-            self.state = DIFFED
+            self.state = State.diffed
             self.diff_map = diff_map
             self.calculate_statistics()
             self.render_diff()
@@ -161,7 +188,7 @@ class DiffHandler(Handler):
                     initialize_highlighter(self.opts.pygments_style)
                 except StyleNotFound as e:
                     self.report_traceback_on_exit = str(e)
-                    self.quit_loop(1)
+                    self.terminate(1)
                     return
                 self.syntax_highlight()
 
@@ -170,13 +197,15 @@ class DiffHandler(Handler):
             self.asyncio_loop.call_soon_threadsafe(diff_done, diff_map)
 
         self.asyncio_loop.run_in_executor(None, diff, self.collection, self.current_context_count)
+        self.doing_background_work = BackgroundWork.diffing
 
     def syntax_highlight(self) -> None:
 
         def highlighting_done(hdata: Union[str, Dict[str, 'DiffHighlight']]) -> None:
+            self.doing_background_work = BackgroundWork.none
             if isinstance(hdata, str):
                 self.report_traceback_on_exit = hdata
-                self.quit_loop(1)
+                self.terminate(1)
                 return
             set_highlight_data(hdata)
             self.render_diff()
@@ -187,6 +216,7 @@ class DiffHandler(Handler):
             self.asyncio_loop.call_soon_threadsafe(highlighting_done, result)
 
         self.asyncio_loop.run_in_executor(None, highlight, self.collection, self.opts.syntax_aliases)
+        self.doing_background_work = BackgroundWork.highlighting
 
     def calculate_statistics(self) -> None:
         self.added_count = self.collection.added_count
@@ -302,7 +332,7 @@ class DiffHandler(Handler):
         self.create_collection()
 
     def enforce_cursor_state(self) -> None:
-        self.cmd.set_cursor_visible(self.state == COMMAND)
+        self.cmd.set_cursor_visible(self.state is State.command)
 
     def draw_lines(self, num: int, offset: int = 0) -> None:
         offset += self.scroll_pos
@@ -317,7 +347,7 @@ class DiffHandler(Handler):
                 text = line.text
                 if line.image_data is not None:
                     image_involved = True
-            self.write('\r\x1b[K' + text + '\x1b[0m')
+            self.write(f'\r\x1b[K{text}\x1b[0m')
             if self.current_search is not None:
                 self.current_search.highlight_line(self.write, lpos)
             if i < num - 1:
@@ -334,8 +364,8 @@ class DiffHandler(Handler):
             if candidate.image.image_id == image_id:
                 q = self.xpos_for_image(row, candidate, is_left)
                 if q is not None:
-                    pl['x'] = q[0]
-                    pl['y'] = row
+                    pl.x = q[0]
+                    pl.y = row
                     return True
             return False
 
@@ -405,7 +435,7 @@ class DiffHandler(Handler):
 
     def draw_screen(self) -> None:
         self.enforce_cursor_state()
-        if self.state < DIFFED:
+        if self.state.value < State.diffed.value:
             self.cmd.clear_screen()
             self.write(_('Calculating diff, please wait...'))
             return
@@ -415,17 +445,17 @@ class DiffHandler(Handler):
         self.draw_status_line()
 
     def draw_status_line(self) -> None:
-        if self.state < DIFFED:
+        if self.state.value < State.diffed.value:
             return
         self.enforce_cursor_state()
         self.cmd.set_cursor_position(0, self.num_lines)
         self.cmd.clear_to_eol()
-        if self.state is COMMAND:
+        if self.state is State.command:
             self.line_edit.write(self.write)
-        elif self.state is MESSAGE:
+        elif self.state is State.message:
             self.cmd.styled(self.message, reverse=True)
         else:
-            sp = '{:.0%}'.format(self.scroll_pos/self.max_scroll_pos) if self.scroll_pos and self.max_scroll_pos else '0%'
+            sp = f'{self.scroll_pos/self.max_scroll_pos:.0%}' if self.scroll_pos and self.max_scroll_pos else '0%'
             scroll_frac = styled(sp, fg=self.opts.margin_fg)
             if self.current_search is None:
                 counts = '{}{}{}'.format(
@@ -434,8 +464,8 @@ class DiffHandler(Handler):
                         styled(str(self.removed_count), fg=self.opts.highlight_removed_bg)
                 )
             else:
-                counts = styled('{} matches'.format(len(self.current_search)), fg=self.opts.margin_fg)
-            suffix = counts + '  ' + scroll_frac
+                counts = styled(f'{len(self.current_search)} matches', fg=self.opts.margin_fg)
+            suffix = f'{counts}  {scroll_frac}'
             prefix = styled(':', fg=self.opts.margin_fg)
             filler = self.screen_size.cols - wcswidth(prefix) - wcswidth(suffix)
             text = '{}{}{}'.format(prefix, ' ' * filler, suffix)
@@ -445,16 +475,16 @@ class DiffHandler(Handler):
         new_ctx = max(0, new_ctx)
         if new_ctx != self.current_context_count:
             self.current_context_count = new_ctx
-            self.state = COLLECTED
+            self.state = State.collected
             self.generate_diff()
             self.restore_position = self.current_position
             self.draw_screen()
 
     def start_search(self, is_regex: bool, is_backward: bool) -> None:
-        if self.state != DIFFED:
+        if self.state is not State.diffed:
             self.cmd.bell()
             return
-        self.state = COMMAND
+        self.state = State.command
         self.line_edit.clear()
         self.line_edit.add_text('?' if is_backward else '/')
         self.current_search_is_regex = is_regex
@@ -468,59 +498,55 @@ class DiffHandler(Handler):
         try:
             self.current_search = Search(self.opts, query[1:], self.current_search_is_regex, query[0] == '?')
         except BadRegex:
-            self.state = MESSAGE
+            self.state = State.message
             self.message = sanitize(_('Bad regex: {}').format(query[1:]))
             self.cmd.bell()
         else:
             if self.current_search(self.diff_lines, self.margin_size, self.screen_size.cols):
                 self.scroll_to_next_match(include_current=True)
             else:
-                self.state = MESSAGE
+                self.state = State.message
                 self.message = sanitize(_('No matches found'))
                 self.cmd.bell()
 
-    def on_text(self, text: str, in_bracketed_paste: bool = False) -> None:
-        if self.state is COMMAND:
-            self.line_edit.on_text(text, in_bracketed_paste)
-            self.draw_status_line()
-            return
-        if self.state is MESSAGE:
-            self.state = DIFFED
-            self.draw_status_line()
-            return
-        action = self.shortcut_action(text)
-        if action is not None:
-            return self.perform_action(action)
-
-    def on_key(self, key_event: KeyEvent) -> None:
-        if self.state is MESSAGE:
-            if key_event.type is not RELEASE:
-                self.state = DIFFED
-                self.draw_status_line()
-            return
-        if self.state is COMMAND:
-            if self.line_edit.on_key(key_event):
-                if not self.line_edit.current_input:
-                    self.state = DIFFED
+    def on_key_event(self, key_event: KeyEvent, in_bracketed_paste: bool = False) -> None:
+        if key_event.text:
+            if self.state is State.command:
+                self.line_edit.on_text(key_event.text, in_bracketed_paste)
                 self.draw_status_line()
                 return
-        if key_event.type is RELEASE:
-            return
-        if self.state is COMMAND:
-            if key_event.key is ESCAPE:
-                self.state = DIFFED
+            if self.state is State.message:
+                self.state = State.diffed
                 self.draw_status_line()
                 return
-            if key_event is enter_key:
-                self.state = DIFFED
-                self.do_search()
-                self.line_edit.clear()
+        else:
+            if self.state is State.message:
+                if key_event.type is not EventType.RELEASE:
+                    self.state = State.diffed
+                    self.draw_status_line()
+                return
+            if self.state is State.command:
+                if self.line_edit.on_key(key_event):
+                    if not self.line_edit.current_input:
+                        self.state = State.diffed
+                    self.draw_status_line()
+                    return
+                if key_event.matches('enter'):
+                    self.state = State.diffed
+                    self.do_search()
+                    self.line_edit.clear()
+                    self.draw_screen()
+                    return
+                if key_event.matches('esc'):
+                    self.state = State.diffed
+                    self.draw_status_line()
+                    return
+            if self.state.value >= State.diffed.value and self.current_search is not None and key_event.matches('esc'):
+                self.current_search = None
                 self.draw_screen()
                 return
-        if self.state >= DIFFED and self.current_search is not None and key_event.key is ESCAPE:
-            self.current_search = None
-            self.draw_screen()
-            return
+            if key_event.type is EventType.RELEASE:
+                return
         action = self.shortcut_action(key_event)
         if action is not None:
             return self.perform_action(action)
@@ -528,16 +554,16 @@ class DiffHandler(Handler):
     def on_resize(self, screen_size: ScreenSize) -> None:
         self.screen_size = screen_size
         self.set_scrolling_region()
-        if self.state > COLLECTED:
+        if self.state.value > State.collected.value:
             self.image_manager.delete_all_sent_images()
             self.render_diff()
         self.draw_screen()
 
     def on_interrupt(self) -> None:
-        self.quit_loop(1)
+        self.terminate(1)
 
     def on_eot(self) -> None:
-        self.quit_loop(1)
+        self.terminate(1)
 
 
 OPTIONS = partial('''\
@@ -566,8 +592,8 @@ class ShowWarning:
     def __init__(self) -> None:
         self.warnings: List[str] = []
 
-    def __call__(self, message: str, category: object, filename: str, lineno: int, file: object = None, line: object = None) -> None:
-        if category is ImageSupportWarning:
+    def __call__(self, message: Any, category: Any, filename: str, lineno: int, file: object = None, line: object = None) -> None:
+        if category is ImageSupportWarning and isinstance(message, str):
             showwarning.warnings.append(message)
 
 
@@ -582,17 +608,39 @@ def terminate_processes(processes: Iterable[int]) -> None:
             os.kill(pid, signal.SIGKILL)
 
 
+def get_ssh_file(hostname: str, rpath: str) -> str:
+    import io
+    import shutil
+    import tarfile
+    tdir = tempfile.mkdtemp(suffix=f'-{hostname}')
+    add_remote_dir(tdir)
+    atexit.register(shutil.rmtree, tdir)
+    is_abs = rpath.startswith('/')
+    rpath = rpath.lstrip('/')
+    cmd = ['ssh', hostname, 'tar', '-c', '-f', '-']
+    if is_abs:
+        cmd.extend(('-C', '/'))
+    cmd.append(rpath)
+    p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
+    assert p.stdout is not None
+    raw = p.stdout.read()
+    if p.wait() != 0:
+        raise SystemExit(p.returncode)
+    with tarfile.open(fileobj=io.BytesIO(raw), mode='r:') as tf:
+        members = tf.getmembers()
+        tf.extractall(tdir)
+        if len(members) == 1:
+            for root, dirs, files in os.walk(tdir):
+                if files:
+                    return os.path.join(root, files[0])
+        return os.path.abspath(os.path.join(tdir, rpath))
+
+
 def get_remote_file(path: str) -> str:
     if path.startswith('ssh:'):
         parts = path.split(':', 2)
         if len(parts) == 3:
-            hostname, rpath = parts[1:]
-            with tempfile.NamedTemporaryFile(suffix='-' + os.path.basename(rpath), prefix='remote:', delete=False) as tf:
-                atexit.register(os.remove, tf.name)
-                p = subprocess.Popen(['ssh', hostname, 'cat', rpath], stdout=tf)
-                if p.wait() != 0:
-                    raise SystemExit(p.returncode)
-                return tf.name
+            return get_ssh_file(parts[1], parts[2])
     return path
 
 
@@ -603,15 +651,15 @@ def main(args: List[str]) -> None:
         raise SystemExit('You must specify exactly two files/directories to compare')
     left, right = items
     global_data.title = _('{} vs. {}').format(left, right)
-    if os.path.isdir(left) != os.path.isdir(right):
-        raise SystemExit('The items to be diffed should both be either directories or files. Comparing a directory to a file is not valid.')
     opts = init_config(cli_opts)
     set_diff_command(opts.diff_cmd)
     lines_for_path.replace_tab_by = opts.replace_tab_by
     left, right = map(get_remote_file, (left, right))
+    if os.path.isdir(left) != os.path.isdir(right):
+        raise SystemExit('The items to be diffed should both be either directories or files. Comparing a directory to a file is not valid.')
     for f in left, right:
         if not os.path.exists(f):
-            raise SystemExit('{} does not exist'.format(f))
+            raise SystemExit(f'{f} does not exist')
 
     loop = Loop()
     handler = DiffHandler(cli_opts, opts, left, right)
@@ -619,9 +667,10 @@ def main(args: List[str]) -> None:
     for message in showwarning.warnings:
         from kitty.utils import safe_print
         safe_print(message, file=sys.stderr)
-    highlight_processes = getattr(highlight_collection, 'processes', ())
-    terminate_processes(tuple(highlight_processes))
-    terminate_processes(tuple(worker_processes))
+    if handler.doing_background_work is BackgroundWork.highlighting:
+        terminate_processes(tuple(get_highlight_processes()))
+    elif handler.doing_background_work == BackgroundWork.diffing:
+        terminate_processes(tuple(worker_processes))
     if loop.return_code != 0:
         if handler.report_traceback_on_exit:
             print(handler.report_traceback_on_exit, file=sys.stderr)
@@ -637,5 +686,5 @@ elif __name__ == '__doc__':
     cd['options'] = OPTIONS
     cd['help_text'] = help_text
 elif __name__ == '__conf__':
-    from .config_data import all_options
-    sys.all_options = all_options  # type: ignore
+    from .options.definition import definition
+    sys.options_definition = definition  # type: ignore
