@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>
 
 import re
@@ -53,7 +52,7 @@ defaults for all users.
 
 def surround(x: str, start: int, end: int) -> str:
     if sys.stdout.isatty():
-        x = '\033[{}m{}\033[{}m'.format(start, x, end)
+        x = f'\033[{start}m{x}\033[{end}m'
     return x
 
 
@@ -123,7 +122,7 @@ def option(x: str) -> str:
 
 @role
 def code(x: str) -> str:
-    return x
+    return cyan(x)
 
 
 @role
@@ -185,7 +184,7 @@ def parse_option_spec(spec: Optional[str] = None) -> Tuple[OptionSpecSeq, Option
                 }
                 state = METADATA
                 continue
-            raise ValueError('Invalid option spec, unexpected line: {}'.format(line))
+            raise ValueError(f'Invalid option spec, unexpected line: {line}')
         elif state is METADATA:
             m = mpat.match(line)
             if m is None:
@@ -227,7 +226,7 @@ def prettify(text: str) -> str:
     def identity(x: str) -> str:
         return x
 
-    def sub(m: Match) -> str:
+    def sub(m: 'Match[str]') -> str:
         role, text = m.group(1, 2)
         return role_map.get(role, identity)(text)
 
@@ -243,7 +242,7 @@ def version(add_rev: bool = False) -> str:
     rev = ''
     from . import fast_data_types
     if add_rev and hasattr(fast_data_types, 'KITTY_VCS_REV'):
-        rev = ' ({})'.format(fast_data_types.KITTY_VCS_REV[:10])
+        rev = f' ({fast_data_types.KITTY_VCS_REV[:10]})'
     return '{} {}{} created by {}'.format(italic(appname), green(str_version), rev, title('Kovid Goyal'))
 
 
@@ -334,7 +333,7 @@ class PrintHelpForSeq:
             a('{}:'.format(title('Options')))
         for opt in seq:
             if isinstance(opt, str):
-                a('{}:'.format(title(opt)))
+                a(f'{title(opt)}:')
                 continue
             help_text = opt['help']
             if help_text == '!':
@@ -347,7 +346,7 @@ class PrintHelpForSeq:
                 t = help_text.replace('%default', str(defval))
                 wa(prettify(t.strip()), indent=4)
                 if defval is not None:
-                    wa('Default: {}'.format(defval), indent=4)
+                    wa(f'Default: {defval}', indent=4)
                 if opt.get('choices'):
                     wa('Choices: {}'.format(', '.join(opt['choices'])), indent=4)
                 a('')
@@ -384,7 +383,7 @@ def seq_as_rst(
     a('.. highlight:: sh')
     a('.. code-block:: sh')
     a('')
-    a('  {} {}{}'.format(appname, optstring, usage))
+    a(f'  {appname} {optstring}{usage}')
     a('')
     message = message or default_msg
     a(prettify_rst(message))
@@ -412,9 +411,9 @@ def seq_as_rst(
             a('')
             a(textwrap.indent(prettify_rst(t), ' ' * 4))
             if defval is not None:
-                a(textwrap.indent('Default: :code:`{}`'.format(defval), ' ' * 4))
+                a(textwrap.indent(f'Default: :code:`{defval}`', ' ' * 4))
             if opt.get('choices'):
-                a(textwrap.indent('Choices: :code:`{}`'.format(', '.join(sorted(opt['choices']))), ' ' * 4))
+                a(textwrap.indent('Choices: {}'.format(', '.join(f':code:`{c}`' for c in sorted(opt['choices']))), ' ' * 4))
             a('')
 
     text = '\n'.join(blocks)
@@ -423,7 +422,7 @@ def seq_as_rst(
 
 def as_type_stub(seq: OptionSpecSeq, disabled: OptionSpecSeq, class_name: str, extra_fields: Sequence[str] = ()) -> str:
     from itertools import chain
-    ans: List[str] = ['class {}:'.format(class_name)]
+    ans: List[str] = [f'class {class_name}:']
     for opt in chain(seq, disabled):
         if isinstance(opt, str):
             continue
@@ -443,10 +442,10 @@ def as_type_stub(seq: OptionSpecSeq, disabled: OptionSpecSeq, class_name: str, e
         elif otype.startswith('bool-'):
             t = 'bool'
         else:
-            raise ValueError('Unknown CLI option type: {}'.format(otype))
-        ans.append('    {}: {}'.format(name, t))
+            raise ValueError(f'Unknown CLI option type: {otype}')
+        ans.append(f'    {name}: {t}')
     for x in extra_fields:
-        ans.append('    {}'.format(x))
+        ans.append(f'    {x}')
     return '\n'.join(ans) + '\n\n\n'
 
 
@@ -485,7 +484,7 @@ class Options:
     def opt_for_alias(self, alias: str) -> OptionDict:
         opt = self.alias_map.get(alias)
         if opt is None:
-            raise SystemExit('Unknown option: {}'.format(emph(alias)))
+            raise SystemExit(f'Unknown option: {emph(alias)}')
         return opt
 
     def needs_arg(self, alias: str) -> bool:
@@ -546,7 +545,7 @@ def parse_cmdline(oc: Options, disabled: OptionSpecSeq, ans: Any, args: Optional
                 needs_arg = oc.needs_arg(parts[0])
                 if not needs_arg:
                     if len(parts) != 1:
-                        raise SystemExit('The {} option does not accept arguments'.format(emph(parts[0])))
+                        raise SystemExit(f'The {emph(parts[0])} option does not accept arguments')
                     oc.process_arg(parts[0])
                     continue
                 if len(parts) == 1:
@@ -561,7 +560,7 @@ def parse_cmdline(oc: Options, disabled: OptionSpecSeq, ans: Any, args: Optional
             oc.process_arg(current_option, arg)
             current_option, state = None, NORMAL
     if state is EXPECTING_ARG:
-        raise SystemExit('An argument is required for the option: {}'.format(emph(arg)))
+        raise SystemExit(f'An argument is required for the option: {emph(arg)}')
 
     for key, val in oc.values_map.items():
         setattr(ans, key, val)
@@ -571,15 +570,6 @@ def parse_cmdline(oc: Options, disabled: OptionSpecSeq, ans: Any, args: Optional
     return leftover_args
 
 
-WATCHER_DEFINITION = '''
---watcher -w
-type=list
-Path to a python file. Appropriately named functions in this file will be called
-for various events, such as when the window is resized, focused or closed. See the section
-on watchers in the launch command documentation :doc:`launch`. Relative paths are
-resolved relative to the kitty config directory.'''
-
-
 def options_spec() -> str:
     if not hasattr(options_spec, 'ans'):
         OPTIONS = '''
@@ -627,11 +617,6 @@ Path to a file containing the startup :italic:`session` (tabs, windows, layout,
 Use - to read from STDIN. See the README file for details and an example.
 
 
-{watcher}
-Note that this watcher will be added only to all initially created windows, not new windows
-created after startup.
-
-
 --hold
 type=bool-set
 Remain open after child process exits. Note that this only affects the first
@@ -672,7 +657,7 @@ respect to the temporary directory. To control kitty, you can send
 it commands with :italic:`kitty @` using the :option:`kitty @ --to` option to
 specify this address. This option will be ignored, unless you set
 :opt:`allow_remote_control` to yes in :file:`kitty.conf`. Note that if you run
-:italic:`kitty @` within a kitty window, there is no need to specify the :italic:`--to`
+:italic:`kitty @` within a kitty window, there is no need to specify the :option:`kitty @ --to`
 option as it is read automatically from the environment. For UNIX sockets, this
 can also be specified in :file:`kitty.conf`.
 
@@ -697,7 +682,7 @@ Output commands received from child process to stdout
 
 
 --replay-commands
-Replay previously dumped commands. Specify the path to a dump file previously created by --dump-commands. You
+Replay previously dumped commands. Specify the path to a dump file previously created by :option:`kitty --dump-commands`. You
 can open a new kitty window to replay the commands with::
 
     kitty sh -c "kitty --replay-commands /path/to/dump/file; read"
@@ -725,13 +710,16 @@ type=bool-set
 Print out information about the selection of fallback fonts for characters not present in the main font.
 
 
+--watcher
+This option is deprecated in favor of the :opt:`watcher` option in kitty.conf and should not be used.
+
+
 --execute -e
 type=bool-set
 !
 '''
         setattr(options_spec, 'ans', OPTIONS.format(
             appname=appname, config_help=CONFIG_HELP.format(appname=appname, conf_name=appname),
-            watcher=WATCHER_DEFINITION
         ))
     ans: str = getattr(options_spec, 'ans')
     return ans
@@ -781,7 +769,9 @@ SYSTEM_CONF = '/etc/xdg/kitty/kitty.conf'
 def create_opts(args: CLIOptions, accumulate_bad_lines: Optional[List[BadLineType]] = None) -> KittyOpts:
     from .config import load_config
     config = tuple(resolve_config(SYSTEM_CONF, defconf, args.config))
-    overrides = (a.replace('=', ' ', 1) for a in args.override or ())
+    # Does not cover the case where `name =` when `=` is the value.
+    pat = re.compile(r'^([a-zA-Z0-9_]+)[ \t]*=')
+    overrides = (pat.sub(r'\1 ', a.lstrip()) for a in args.override or ())
     opts = load_config(*config, overrides=overrides, accumulate_bad_lines=accumulate_bad_lines)
     return opts
 
