@@ -1,22 +1,21 @@
 #!/usr/bin/env python
-# vim:fileencoding=utf-8
 # License: GPLv3 Copyright: 2020, Kovid Goyal <kovid at kovidgoyal.net>
 
 from functools import partial
 from itertools import repeat
 from typing import (
-    Dict, Generator, Iterable, Iterator, List, NamedTuple, Optional, Sequence,
-    Tuple
+    Any, Dict, Generator, Iterable, Iterator, List, NamedTuple, Optional,
+    Sequence, Tuple
 )
 
-from kitty.constants import Edges, WindowGeometry
+from kitty.borders import BorderColor
 from kitty.fast_data_types import (
     Region, set_active_window, viewport_for_window
 )
-from kitty.options_stub import Options
+from kitty.options.types import Options
+from kitty.types import Edges, WindowGeometry
 from kitty.typing import TypedDict, WindowType
 from kitty.window_list import WindowGroup, WindowList
-from kitty.borders import BorderColor
 
 
 class BorderLine(NamedTuple):
@@ -29,6 +28,9 @@ class LayoutOpts:
     def __init__(self, data: Dict[str, str]):
         pass
 
+    def serialized(self) -> Dict[str, Any]:
+        return {}
+
 
 class LayoutData(NamedTuple):
     content_pos: int = 0
@@ -67,6 +69,7 @@ def idx_for_id(win_id: int, windows: Iterable[WindowType]) -> Optional[int]:
     for i, w in enumerate(windows):
         if w.id == win_id:
             return i
+    return None
 
 
 def set_layout_options(opts: Options) -> None:
@@ -214,7 +217,7 @@ class Layout:
         self.blank_rects: List[Rect] = []
         self.layout_opts = self.parse_layout_opts(layout_opts)
         assert self.name is not None
-        self.full_name = self.name + ((':' + layout_opts) if layout_opts else '')
+        self.full_name = f'{self.name}:{layout_opts}' if layout_opts else self.name
         self.remove_all_biases()
 
     def bias_increment_for_cell(self, is_horizontal: bool) -> float:
@@ -374,9 +377,16 @@ class Layout:
     def compute_needs_borders_map(self, all_windows: WindowList) -> Dict[int, bool]:
         return all_windows.compute_needs_borders_map(lgd.draw_active_borders)
 
+    def get_minimal_borders(self, windows: WindowList) -> Generator[BorderLine, None, None]:
+        self._set_dimensions()
+        yield from self.minimal_borders(windows)
+
     def minimal_borders(self, windows: WindowList) -> Generator[BorderLine, None, None]:
         return
         yield BorderLine()  # type: ignore
 
     def layout_action(self, action_name: str, args: Sequence[str], all_windows: WindowList) -> Optional[bool]:
         pass
+
+    def layout_state(self) -> Dict[str, Any]:
+        return {}
