@@ -1,5 +1,4 @@
 #!/usr/bin/env python
-# vim:fileencoding=utf-8
 # License: GPLv3 Copyright: 2020, Kovid Goyal <kovid at kovidgoyal.net>
 
 from functools import lru_cache
@@ -65,7 +64,6 @@ class Grid(Layout):
         return layout_func(num_windows, bias=variable_bias(num_windows, biased_map) if num_windows > 1 else None)
 
     def apply_bias(self, idx: int, increment: float, all_windows: WindowList, is_horizontal: bool = True) -> bool:
-        b = self.biased_cols if is_horizontal else self.biased_rows
         num_windows = all_windows.num_groups
         ncols, nrows, special_rows, special_col = calc_grid_size(num_windows)
 
@@ -82,6 +80,7 @@ class Grid(Layout):
                 if idx == window_idx:
                     return row_num, col_num
                 row_num += 1
+            return 0, 0
 
         row_num, col_num = position_for_window_idx(idx)
 
