@@ -1,9 +1,9 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import json
 import os
+import re
 import sys
 import weakref
 from collections import deque
@@ -11,9 +11,10 @@ from enum import IntEnum
 from functools import partial
 from gettext import gettext as _
 from itertools import chain
+from time import monotonic
 from typing import (
-    Any, Callable, Deque, Dict, Iterable, List, NamedTuple, Optional, Pattern,
-    Sequence, Tuple, Union
+    TYPE_CHECKING, Any, Callable, Deque, Dict, Iterable, List, NamedTuple,
+    Optional, Pattern, Sequence, Tuple, Union
 )
 
 from .child import ProcessDesc
@@ -23,31 +24,49 @@ from .constants import appname, is_macos, wakeup
 from .fast_data_types import (
     BGIMAGE_PROGRAM, BLIT_PROGRAM, CELL_BG_PROGRAM, CELL_FG_PROGRAM,
     CELL_PROGRAM, CELL_SPECIAL_PROGRAM, CURSOR_BEAM, CURSOR_BLOCK,
-    CURSOR_UNDERLINE, DCS, DECORATION, DIM, GLFW_MOD_CONTROL,
+    CURSOR_UNDERLINE, DCS, DECORATION, DECORATION_MASK, DIM, GLFW_MOD_CONTROL,
     GRAPHICS_ALPHA_MASK_PROGRAM, GRAPHICS_PREMULT_PROGRAM, GRAPHICS_PROGRAM,
-    MARK, MARK_MASK, NO_CURSOR_SHAPE, OSC, REVERSE, SCROLL_FULL, SCROLL_LINE,
-    SCROLL_PAGE, STRIKETHROUGH, TINT_PROGRAM, KeyEvent, Screen, add_timer,
-    add_window, cell_size_for_window, click_mouse_url, compile_program,
-    encode_key_for_tty, get_boss, get_clipboard_string, get_options,
-    init_cell_program, mouse_selection, pt_to_px, set_clipboard_string,
-    set_titlebar_color, set_window_padding, set_window_render_data,
-    update_window_title, update_window_visibility, viewport_for_window
+    MARK, MARK_MASK, NO_CURSOR_SHAPE, NUM_UNDERLINE_STYLES, OSC, REVERSE,
+    SCROLL_FULL, SCROLL_LINE, SCROLL_PAGE, STRIKETHROUGH, TINT_PROGRAM, Color,
+    KeyEvent, Screen, add_timer, add_window, cell_size_for_window,
+    click_mouse_cmd_output, click_mouse_url, compile_program,
+    current_os_window, encode_key_for_tty, get_boss, get_click_interval,
+    get_clipboard_string, get_options, init_cell_program, mark_os_window_dirty,
+    mouse_selection, move_cursor_to_mouse_if_in_prompt, pt_to_px,
+    set_clipboard_string, set_titlebar_color, set_window_logo,
+    set_window_padding, set_window_render_data, update_ime_position_for_window,
+    update_window_title, update_window_visibility
 )
 from .keys import keyboard_mode_name, mod_mask
 from .notify import NotificationCommand, handle_notification_cmd
 from .options.types import Options
 from .rgb import to_color
 from .terminfo import get_capabilities
-from .types import MouseEvent, ScreenGeometry, WindowGeometry, ac
+from .types import MouseEvent, WindowGeometry, ac
 from .typing import BossType, ChildType, EdgeLiteral, TabType, TypedDict
 from .utils import (
-    color_as_int, get_primary_selection, load_shaders, log_error, open_cmd,
-    open_url, parse_color_set, sanitize_title, set_primary_selection
+    get_primary_selection, kitty_ansi_sanitizer_pat, load_shaders, log_error,
+    open_cmd, open_url, parse_color_set, resolve_custom_file, sanitize_title,
+    set_primary_selection
 )
 
 MatchPatternType = Union[Pattern[str], Tuple[Pattern[str], Optional[Pattern[str]]]]
 
 
+if TYPE_CHECKING:
+    from .file_transmission import FileTransmission
+
+
+def process_title_from_child(title: str, is_base64: bool) -> str:
+    if is_base64:
+        from base64 import standard_b64decode
+        try:
+            title = standard_b64decode(title).decode('utf-8', 'replace')
+        except Exception:
+            title = 'undecodeable title'
+    return sanitize_title(title)
+
+
 class WindowDict(TypedDict):
     id: int
     is_focused: bool
@@ -82,6 +101,10 @@ class DynamicColor(IntEnum):
     default_fg, default_bg, cursor_color, highlight_fg, highlight_bg = range(1, 6)
 
 
+class CommandOutput(IntEnum):
+    last_run, first_on_screen, last_visited = 0, 1, 2
+
+
 DYNAMIC_COLOR_CODES = {
     10: DynamicColor.default_fg,
     11: DynamicColor.default_bg,
@@ -125,7 +148,7 @@ class Watchers:
         ans = Watchers()
         ans.on_close = self.on_close[:]
         ans.on_resize = self.on_resize[:]
-        ans.on_focus_change = self.on_focus_change
+        ans.on_focus_change = self.on_focus_change[:]
         return ans
 
     @property
@@ -144,13 +167,12 @@ def call_watchers(windowref: Callable[[], Optional['Window']], which: str, data:
     add_timer(callback, 0, False)
 
 
-def calculate_gl_geometry(window_geometry: WindowGeometry, viewport_width: int, viewport_height: int, cell_width: int, cell_height: int) -> ScreenGeometry:
-    dx, dy = 2 * cell_width / viewport_width, 2 * cell_height / viewport_height
-    xmargin = window_geometry.left / viewport_width
-    ymargin = window_geometry.top / viewport_height
-    xstart = -1 + 2 * xmargin
-    ystart = 1 - 2 * ymargin
-    return ScreenGeometry(xstart, ystart, window_geometry.xnum, window_geometry.ynum, dx, dy)
+def pagerhist(screen: Screen, as_ansi: bool = False, add_wrap_markers: bool = True, upto_output_start: bool = False) -> str:
+    pht = screen.historybuf.pagerhist_as_text(upto_output_start)
+    if pht and (not as_ansi or not add_wrap_markers):
+        sanitizer = text_sanitizer(as_ansi, add_wrap_markers)
+        pht = sanitizer(pht)
+    return pht
 
 
 def as_text(
@@ -182,13 +204,8 @@ def as_text(
             ctext += f'\x1b[{code} q'
 
     if add_history:
-        h: List[str] = []
-        pht = screen.historybuf.pagerhist_as_text()
-        if pht:
-            h.append(pht)
-        if h and (not as_ansi or not add_wrap_markers):
-            sanitizer = text_sanitizer(as_ansi, add_wrap_markers)
-            h = list(map(sanitizer, h))
+        pht = pagerhist(screen, as_ansi, add_wrap_markers)
+        h: List[str] = [pht] if pht else []
         screen.historybuf.as_text(h.append, as_ansi, add_wrap_markers)
         if h:
             if not screen.linebuf.is_continued(0):
@@ -205,9 +222,16 @@ def as_text(
     return ans
 
 
-class LoadShaderPrograms:
+def multi_replace(src: str, **replacements: Any) -> str:
+    r = {k: str(v) for k, v in replacements.items()}
 
-    use_selection_fg = True
+    def sub(m: 're.Match[str]') -> str:
+        return r.get(m.group(1), m.group(1))
+
+    return re.sub(r'\{([A-Z_]+)\}', sub, src)
+
+
+class LoadShaderPrograms:
 
     def __call__(self, semi_transparent: bool = False) -> None:
         compile_program(BLIT_PROGRAM, *load_shaders('blit'))
@@ -219,22 +243,22 @@ class LoadShaderPrograms:
                 'SPECIAL': CELL_SPECIAL_PROGRAM,
                 'FOREGROUND': CELL_FG_PROGRAM,
         }.items():
-            vv, ff = v.replace('WHICH_PROGRAM', which), f.replace('WHICH_PROGRAM', which)
-            for gln, pyn in {
-                    'REVERSE_SHIFT': REVERSE,
-                    'STRIKE_SHIFT': STRIKETHROUGH,
-                    'DIM_SHIFT': DIM,
-                    'DECORATION_SHIFT': DECORATION,
-                    'MARK_SHIFT': MARK,
-                    'MARK_MASK': MARK_MASK,
-            }.items():
-                vv = vv.replace('{{{}}}'.format(gln), str(pyn), 1)
+            ff = f.replace('{WHICH_PROGRAM}', which)
+            vv = multi_replace(
+                v,
+                WHICH_PROGRAM=which,
+                REVERSE_SHIFT=REVERSE,
+                STRIKE_SHIFT=STRIKETHROUGH,
+                DIM_SHIFT=DIM,
+                DECORATION_SHIFT=DECORATION,
+                MARK_SHIFT=MARK,
+                MARK_MASK=MARK_MASK,
+                DECORATION_MASK=DECORATION_MASK,
+                STRIKE_SPRITE_INDEX=NUM_UNDERLINE_STYLES + 1,
+            )
             if semi_transparent:
                 vv = vv.replace('#define NOT_TRANSPARENT', '#define TRANSPARENT')
                 ff = ff.replace('#define NOT_TRANSPARENT', '#define TRANSPARENT')
-            if not load_shader_programs.use_selection_fg:
-                vv = vv.replace('#define USE_SELECTION_FG', '#define DONT_USE_SELECTION_FG')
-                ff = ff.replace('#define USE_SELECTION_FG', '#define DONT_USE_SELECTION_FG')
             compile_program(p, vv, ff)
 
         v, f = load_shaders('graphics')
@@ -258,23 +282,19 @@ load_shader_programs = LoadShaderPrograms()
 
 def setup_colors(screen: Screen, opts: Options) -> None:
     screen.color_profile.update_ansi_color_table(build_ansi_color_table(opts))
-    cursor_text_color = opts.cursor_text_color or (12, 12, 12)
-    cursor_text_color_as_bg = 3 if opts.cursor_text_color is None else 1
-    sfg = (0, 0, 0) if opts.selection_foreground is None else opts.selection_foreground
-    screen.color_profile.set_configured_colors(*map(color_as_int, (
-        opts.foreground, opts.background, opts.cursor,
-        cursor_text_color, (0, 0, cursor_text_color_as_bg),
-        sfg, opts.selection_background)
-    ))
+
+    def s(c: Optional[Color]) -> int:
+        return 0 if c is None else (0xff000000 | int(c))
+    screen.color_profile.set_configured_colors(
+        s(opts.foreground), s(opts.background),
+        s(opts.cursor), s(opts.cursor_text_color),
+        s(opts.selection_foreground), s(opts.selection_background),
+        s(opts.visual_bell_color)
+    )
 
 
 def text_sanitizer(as_ansi: bool, add_wrap_markers: bool) -> Callable[[str], str]:
-    pat = getattr(text_sanitizer, 'pat', None)
-    if pat is None:
-        import re
-        pat = re.compile('\033\\[.*?m')
-        setattr(text_sanitizer, 'pat', pat)
-
+    pat = kitty_ansi_sanitizer_pat()
     ansi, wrap_markers = not as_ansi, not add_wrap_markers
 
     def remove_wrap_markers(line: str) -> str:
@@ -291,6 +311,23 @@ def text_sanitizer(as_ansi: bool, add_wrap_markers: bool) -> Callable[[str], str
     return remove_wrap_markers
 
 
+def cmd_output(screen: Screen, which: CommandOutput = CommandOutput.last_run, as_ansi: bool = False, add_wrap_markers: bool = False) -> str:
+    lines: List[str] = []
+    search_in_pager_hist = screen.cmd_output(which, lines.append, as_ansi, add_wrap_markers)
+    if search_in_pager_hist:
+        pht = pagerhist(screen, as_ansi, add_wrap_markers, True)
+        if pht:
+            lines.insert(0, pht)
+    return ''.join(lines)
+
+
+def process_remote_print(msg: bytes) -> str:
+    from base64 import standard_b64decode
+    from .cli import green
+    text = standard_b64decode(msg).decode('utf-8', 'replace')
+    return text.replace('\x1b', green(r'\e')).replace('\a', green(r'\a')).replace('\0', green(r'\0'))
+
+
 class EdgeWidths:
     left: Optional[float]
     top: Optional[float]
@@ -310,6 +347,32 @@ class EdgeWidths:
         return {'left': self.left, 'right': self.right, 'top': self.top, 'bottom': self.bottom}
 
 
+class GlobalWatchers:
+
+    def __init__(self) -> None:
+        self.options_spec: Optional[Dict[str, str]] = None
+        self.ans = Watchers()
+        self.extra = ''
+
+    def __call__(self) -> Watchers:
+        spec = get_options().watcher
+        if spec == self.options_spec:
+            return self.ans
+        from .launch import load_watch_modules
+        if self.extra:
+            spec = spec.copy()
+            spec[self.extra] = self.extra
+        self.ans = load_watch_modules(spec.keys()) or self.ans
+        self.options_spec = spec.copy()
+        return self.ans
+
+    def set_extra(self, extra: str) -> None:
+        self.extra = extra
+
+
+global_watchers = GlobalWatchers()
+
+
 class Window:
 
     def __init__(
@@ -321,11 +384,18 @@ class Window:
         copy_colors_from: Optional['Window'] = None,
         watchers: Optional[Watchers] = None
     ):
-        self.watchers = watchers or Watchers()
+        if watchers:
+            self.watchers = watchers
+            self.watchers.add(global_watchers())
+        else:
+            self.watchers = global_watchers().copy()
+        self.last_focused_at = 0.
+        self.started_at = monotonic()
         self.current_mouse_event_button = 0
+        self.current_clipboard_read_ask: Optional[bool] = None
         self.prev_osc99_cmd = NotificationCommand()
-        self.action_on_close: Optional[Callable] = None
-        self.action_on_removal: Optional[Callable] = None
+        self.actions_on_close: List[Callable[['Window'], None]] = []
+        self.actions_on_removal: List[Callable[['Window'], None]] = []
         self.current_marker_spec: Optional[Tuple[str, Union[str, Tuple[Tuple[int, str], ...]]]] = None
         self.pty_resized_once = False
         self.last_reported_pty_size = (-1, -1, -1, -1)
@@ -339,7 +409,7 @@ class Window:
         self.margin = EdgeWidths()
         self.padding = EdgeWidths()
         if not self.id:
-            raise Exception('No tab with id: {} in OS Window: {} was found, or the window counter wrapped'.format(tab.id, tab.os_window_id))
+            raise Exception(f'No tab with id: {tab.id} in OS Window: {tab.os_window_id} was found, or the window counter wrapped')
         self.tab_id = tab.id
         self.os_window_id = tab.os_window_id
         self.tabref: Callable[[], Optional[TabType]] = weakref.ref(tab)
@@ -357,6 +427,14 @@ class Window:
         else:
             setup_colors(self.screen, opts)
 
+    @property
+    def file_transmission_control(self) -> 'FileTransmission':
+        ans: Optional['FileTransmission'] = getattr(self, '_file_transmission', None)
+        if ans is None:
+            from .file_transmission import FileTransmission
+            ans = self._file_transmission = FileTransmission(self.id)
+        return ans
+
     def on_dpi_change(self, font_sz: float) -> None:
         self.update_effective_padding()
 
@@ -415,8 +493,7 @@ class Window:
         return self.override_title or self.child_title
 
     def __repr__(self) -> str:
-        return 'Window(title={}, id={})'.format(
-                self.title, self.id)
+        return f'Window(title={self.title}, id={self.id})'
 
     def as_dict(self, is_focused: bool = False, is_self: bool = False) -> WindowDict:
         return dict(
@@ -450,9 +527,17 @@ class Window:
         }
 
     @property
-    def current_colors(self) -> Dict:
+    def current_colors(self) -> Dict[str, Optional[int]]:
         return self.screen.color_profile.as_dict()
 
+    @property
+    def at_prompt(self) -> bool:
+        return self.screen.cursor_at_prompt()
+
+    @property
+    def has_running_program(self) -> bool:
+        return not self.at_prompt
+
     def matches(self, field: str, pat: MatchPatternType) -> bool:
         if not pat:
             return False
@@ -493,46 +578,44 @@ class Window:
         self.screen.mark_as_dirty()
         wakeup()
 
-    def update_position(self, window_geometry: WindowGeometry) -> ScreenGeometry:
-        central, tab_bar, vw, vh, cw, ch = viewport_for_window(self.os_window_id)
-        self.screen_geometry = sg = calculate_gl_geometry(window_geometry, vw, vh, cw, ch)
-        return sg
-
     def set_geometry(self, new_geometry: WindowGeometry) -> None:
         if self.destroyed:
             return
         if self.needs_layout or new_geometry.xnum != self.screen.columns or new_geometry.ynum != self.screen.lines:
             self.screen.resize(new_geometry.ynum, new_geometry.xnum)
-            sg = self.update_position(new_geometry)
             self.needs_layout = False
             call_watchers(weakref.ref(self), 'on_resize', {'old_geometry': self.geometry, 'new_geometry': new_geometry})
-        else:
-            sg = self.update_position(new_geometry)
         current_pty_size = (
             self.screen.lines, self.screen.columns,
             max(0, new_geometry.right - new_geometry.left), max(0, new_geometry.bottom - new_geometry.top))
+        update_ime_position = False
         if current_pty_size != self.last_reported_pty_size:
             get_boss().child_monitor.resize_pty(self.id, *current_pty_size)
             if not self.pty_resized_once:
                 self.pty_resized_once = True
                 self.child.mark_terminal_ready()
+                update_ime_position = True
             self.last_reported_pty_size = current_pty_size
+        else:
+            mark_os_window_dirty(self.os_window_id)
 
         self.geometry = g = new_geometry
-        set_window_render_data(self.os_window_id, self.tab_id, self.id, sg.xstart, sg.ystart, sg.dx, sg.dy, self.screen, *g[:4])
+        set_window_render_data(self.os_window_id, self.tab_id, self.id, self.screen, *g[:4])
         self.update_effective_padding()
+        if update_ime_position:
+            update_ime_position_for_window(self.id, True)
 
     def contains(self, x: int, y: int) -> bool:
         g = self.geometry
         return g.left <= x <= g.right and g.top <= y <= g.bottom
 
     def close(self) -> None:
-        get_boss().close_window(self)
+        get_boss().mark_window_for_close(self)
 
     @ac('misc', '''
         Send the specified text to the active window
 
-        For details, see :sc:`send_text`.
+        See :sc:`send_text <send_text>` for details.
         ''')
     def send_text(self, *args: str) -> bool:
         mode = keyboard_mode_name(self.screen)
@@ -543,6 +626,14 @@ class Window:
         if not text:
             return True
         self.write_to_child(text)
+        return False
+
+    @ac('debug', 'Show a dump of the current lines in the scrollback + screen with their line attributes')
+    def dump_lines_with_attrs(self) -> None:
+        strings: List[str] = []
+        self.screen.dump_lines_with_attrs(strings.append)
+        text = ''.join(strings)
+        get_boss().display_scrollback(self, text, title='Dump of lines', report_cursor=False)
 
     def write_to_child(self, data: Union[str, bytes]) -> None:
         if data:
@@ -562,6 +653,11 @@ class Window:
         self.title_updated()
 
     def desktop_notify(self, osc_code: int, raw_data: str) -> None:
+        if osc_code == 777:
+            if not raw_data.startswith('notify;'):
+                log_error(f'Ignoring unknown OSC 777: {raw_data}')
+                return  # unknown OSC 777
+            raw_data = raw_data[len('notify;'):]
         cmd = handle_notification_cmd(osc_code, raw_data, self.id, self.prev_osc99_cmd)
         if cmd is not None and osc_code == 99:
             self.prev_osc99_cmd = cmd
@@ -577,7 +673,7 @@ class Window:
         action = get_options().mousemap.get(ev)
         if action is None:
             return False
-        return get_boss().dispatch_action(action, window_for_dispatch=self, dispatch_type='MouseEvent')
+        return get_boss().combine(action, window_for_dispatch=self, dispatch_type='MouseEvent')
 
     def open_url(self, url: str, hyperlink_id: int, cwd: Optional[str] = None) -> None:
         opts = get_options()
@@ -603,19 +699,16 @@ class Window:
                     url = urlunparse(purl._replace(netloc=''))
             if opts.allow_hyperlinks & 0b10:
                 from kittens.tui.operations import styled
-                get_boss()._run_kitten('ask', ['--type=choices', '--message', _(
-                    'What would you like to do with this URL:\n') +
-                    styled(unquote(url), fg='yellow'),
-                    '--choice=o:Open', '--choice=c:Copy to clipboard', '--choice=n;red:Nothing'
-                    ],
+                get_boss().choose(
+                    'What would you like to do with this URL:\n' + styled(unquote(url), fg='yellow'),
+                    partial(self.hyperlink_open_confirmed, url, cwd),
+                    'o:Open', 'c:Copy to clipboard', 'n;red:Nothing', default='o',
                     window=self,
-                    custom_callback=partial(self.hyperlink_open_confirmed, url, cwd)
                 )
                 return
         get_boss().open_url(url, cwd=cwd)
 
-    def hyperlink_open_confirmed(self, url: str, cwd: Optional[str], data: Dict[str, Any], *a: Any) -> None:
-        q = data['response']
+    def hyperlink_open_confirmed(self, url: str, cwd: Optional[str], q: str) -> None:
         if q == 'o':
             get_boss().open_url(url, cwd=cwd)
         elif q == 'c':
@@ -624,7 +717,7 @@ class Window:
     def handle_remote_file(self, netloc: str, remote_path: str) -> None:
         from kittens.ssh.main import get_connection_data
         args = self.child.foreground_cmdline
-        conn_data = get_connection_data(args)
+        conn_data = get_connection_data(args, self.child.foreground_cwd or self.child.current_cwd or '')
         if conn_data is None:
             get_boss().show_error('Could not handle remote file', 'No SSH connection data found in: {args}')
             return
@@ -639,15 +732,20 @@ class Window:
         call_watchers(weakref.ref(self), 'on_focus_change', {'focused': focused})
         self.screen.focus_changed(focused)
         if focused:
+            self.last_focused_at = monotonic()
+            update_ime_position_for_window(self.id)
             changed = self.needs_attention
             self.needs_attention = False
             if changed:
                 tab = self.tabref()
                 if tab is not None:
                     tab.relayout_borders()
+        elif self.os_window_id == current_os_window():
+            # Cancel IME composition after loses focus
+            update_ime_position_for_window(self.id, False, True)
 
-    def title_changed(self, new_title: Optional[str]) -> None:
-        self.child_title = sanitize_title(new_title or self.default_title)
+    def title_changed(self, new_title: Optional[str], is_base64: bool = False) -> None:
+        self.child_title = process_title_from_child(new_title or self.default_title, is_base64)
         if self.override_title is None:
             self.title_updated()
 
@@ -686,28 +784,28 @@ class Window:
     def change_titlebar_color(self) -> None:
         opts = get_options()
         val = opts.macos_titlebar_color if is_macos else opts.wayland_titlebar_color
-        if val:
+        if val > 0:
             if (val & 0xff) == 1:
                 val = self.screen.color_profile.default_bg
             else:
                 val = val >> 8
             set_titlebar_color(self.os_window_id, val)
         else:
-            set_titlebar_color(self.os_window_id, 0, True)
+            set_titlebar_color(self.os_window_id, 0, True, -val)
 
     def change_colors(self, changes: Dict[DynamicColor, Optional[str]]) -> None:
         dirtied = default_bg_changed = False
 
-        def item(raw: Optional[str]) -> Optional[int]:
+        def item(raw: Optional[str]) -> int:
             if raw is None:
                 return 0
-            val = to_color(raw)
-            return None if val is None else (color_as_int(val) << 8) | 2
+            v = to_color(raw)
+            if v is None:
+                return 0
+            return 0xff000000 | int(v)
 
         for which, val_ in changes.items():
             val = item(val_)
-            if val is None:
-                continue
             dirtied = True
             setattr(self.screen.color_profile, which.name, val)
             if which.name == 'default_bg':
@@ -721,7 +819,7 @@ class Window:
         r |= r << 8
         g |= g << 8
         b |= b << 8
-        self.screen.send_escape_code_to_child(OSC, '{};rgb:{:04x}/{:04x}/{:04x}'.format(code, r, g, b))
+        self.screen.send_escape_code_to_child(OSC, f'{code};rgb:{r:04x}/{g:04x}/{b:04x}')
 
     def report_notification_activated(self, identifier: str) -> None:
         self.screen.send_escape_code_to_child(OSC, f'99;i={identifier};')
@@ -749,7 +847,9 @@ class Window:
             changed = False
             for c, val in parse_color_set(value):
                 if val is None:  # color query
-                    self.report_color('4;{}'.format(c), *self.screen.color_profile.as_color((c << 8) | 1))
+                    qc = self.screen.color_profile.as_color((c << 8) | 1)
+                    assert qc is not None
+                    self.report_color(f'4;{c}', qc.red, qc.green, qc.blue)
                 else:
                     changed = True
                     cp.set_color(c, val)
@@ -776,14 +876,16 @@ class Window:
         get_boss().handle_remote_cmd(cmd, self)
 
     def handle_remote_print(self, msg: bytes) -> None:
-        from base64 import standard_b64decode
-        text = standard_b64decode(msg).decode('utf-8')
+        text = process_remote_print(msg)
         print(text, end='', file=sys.stderr)
         sys.stderr.flush()
 
     def send_cmd_response(self, response: Any) -> None:
         self.screen.send_escape_code_to_child(DCS, '@kitty-cmd' + json.dumps(response))
 
+    def file_transmission(self, data: str) -> None:
+        self.file_transmission_control.handle_serialized_command(data)
+
     def clipboard_control(self, data: str, is_partial: bool = False) -> None:
         where, text = data.partition(';')[::2]
         if is_partial:
@@ -791,7 +893,8 @@ class Window:
                 self.clipboard_pending = ClipboardPending(where, text)
             else:
                 self.clipboard_pending = self.clipboard_pending._replace(data=self.clipboard_pending[1] + text)
-                if len(self.clipboard_pending.data) > 8 * 1024 * 1024:
+                limit = get_options().clipboard_max_size
+                if limit and len(self.clipboard_pending.data) > limit * 1024 * 1024:
                     log_error('Discarding part of too large OSC 52 paste request')
                     self.clipboard_pending = self.clipboard_pending._replace(data='', truncated=True)
             return
@@ -811,15 +914,17 @@ class Window:
         if text == '?':
             response = None
             if 's' in where or 'c' in where:
+                if 'read-clipboard-ask' in cc:
+                    return self.ask_to_read_clipboard(False)
                 response = get_clipboard_string() if 'read-clipboard' in cc else ''
                 loc = 'c'
             elif 'p' in where:
+                if 'read-primary-ask' in cc:
+                    return self.ask_to_read_clipboard(True)
                 response = get_primary_selection() if 'read-primary' in cc else ''
                 loc = 'p'
             response = response or ''
-            from base64 import standard_b64encode
-            self.screen.send_escape_code_to_child(OSC, '52;{};{}'.format(
-                loc, standard_b64encode(response.encode('utf-8')).decode('ascii')))
+            self.send_osc52(loc, response or '')
 
         else:
             from base64 import standard_b64decode
@@ -836,6 +941,32 @@ class Window:
                     set_primary_selection(text)
         self.clipboard_pending = None
 
+    def send_osc52(self, loc: str, response: str) -> None:
+        from base64 import standard_b64encode
+        self.screen.send_escape_code_to_child(OSC, '52;{};{}'.format(
+            loc, standard_b64encode(response.encode('utf-8')).decode('ascii')))
+
+    def ask_to_read_clipboard(self, primary: bool = False) -> None:
+        if self.current_clipboard_read_ask is not None:
+            self.current_clipboard_read_ask = primary
+            return
+        self.current_clipboard_read_ask = primary
+        get_boss().confirm(_(
+            'A program running in this window wants to read from the system clipboard.'
+            ' Allow it do so, once?'),
+            self.handle_clipboard_confirmation, window=self,
+        )
+
+    def handle_clipboard_confirmation(self, confirmed: bool) -> None:
+        try:
+            loc = 'p' if self.current_clipboard_read_ask else 'c'
+            response = ''
+            if confirmed:
+                response = get_primary_selection() if self.current_clipboard_read_ask else get_clipboard_string()
+            self.send_osc52(loc, response)
+        finally:
+            self.current_clipboard_read_ask = None
+
     def manipulate_title_stack(self, pop: bool, title: str, icon: Any) -> None:
         if title:
             if pop:
@@ -848,14 +979,44 @@ class Window:
     # }}}
 
     # mouse actions {{{
+    @ac('mouse', '''
+        Handle a mouse click
+
+        Try to perform the specified actions one after the other till one of them is successful.
+        Supported actions are::
+
+            selection - check for a selection and if one exists abort processing
+            link - if a link exists under the mouse, click it
+            prompt - if the mouse click happens at a shell prompt move the cursor to the mouse location
+
+        For examples, see :ref:`conf-kitty-mouse.mousemap`
+        ''')
+    def mouse_handle_click(self, *actions: str) -> None:
+        for a in actions:
+            if a == 'selection':
+                if self.screen.has_selection():
+                    break
+            if a == 'link':
+                if click_mouse_url(self.os_window_id, self.tab_id, self.id):
+                    break
+            if a == 'prompt':
+                # Do not send move cursor events too soon after the window is
+                # focused, this is because there are people that click on
+                # windows and start typing immediately and the cursor event
+                # can interfere with that. See https://github.com/kovidgoyal/kitty/issues/4128
+                if monotonic() - self.last_focused_at < 1.5 * get_click_interval():
+                    return
+                if move_cursor_to_mouse_if_in_prompt(self.os_window_id, self.tab_id, self.id):
+                    self.screen.ignore_bells_for(1)
+                    break
+
     @ac('mouse', 'Click the URL under the mouse')
     def mouse_click_url(self) -> None:
-        click_mouse_url(self.os_window_id, self.tab_id, self.id)
+        self.mouse_handle_click('link')
 
     @ac('mouse', 'Click the URL under the mouse only if the screen has no selection')
     def mouse_click_url_or_select(self) -> None:
-        if not self.screen.has_selection():
-            self.mouse_click_url()
+        self.mouse_handle_click('selection', 'link')
 
     @ac('mouse', '''
         Manipulate the selection based on the current mouse position
@@ -876,14 +1037,29 @@ class Window:
         txt = get_boss().current_primary_selection_or_clipboard()
         if txt:
             self.paste(txt)
+
+    @ac('mouse', '''
+        Select clicked command output
+
+        Requires :ref:`shell_integration` to work
+        ''')
+    def mouse_select_command_output(self) -> None:
+        click_mouse_cmd_output(self.os_window_id, self.tab_id, self.id, True)
+
+    @ac('mouse', '''
+        Show clicked command output in a pager like less
+
+        Requires :ref:`shell_integration` to work
+        ''')
+    def mouse_show_command_output(self) -> None:
+        if click_mouse_cmd_output(self.os_window_id, self.tab_id, self.id, False):
+            self.show_cmd_output(CommandOutput.last_visited, 'Clicked command output')
     # }}}
 
-    def text_for_selection(self) -> str:
-        lines = self.screen.text_for_selection()
+    def text_for_selection(self, as_ansi: bool = False) -> str:
         sts = get_options().strip_trailing_spaces
-        if sts == 'always' or (
-                sts == 'smart' and not self.screen.is_rectangle_select()):
-            return ''.join((ln.rstrip() or '\n') for ln in lines)
+        strip_trailing_spaces = sts == 'always' or (sts == 'smart' and not self.screen.is_rectangle_select())
+        lines = self.screen.text_for_selection(as_ansi, strip_trailing_spaces)
         return ''.join(lines)
 
     def call_watchers(self, which: Iterable[Watcher], data: Dict[str, Any]) -> None:
@@ -899,6 +1075,9 @@ class Window:
         self.call_watchers(self.watchers.on_close, {})
         self.destroyed = True
         if hasattr(self, 'screen'):
+            if self.is_active and self.os_window_id == current_os_window():
+                # Cancel IME composition when window is destroyed
+                update_ime_position_for_window(self.id, False, True)
             # Remove cycles so that screen is de-allocated immediately
             self.screen.reset_callbacks()
             del self.screen
@@ -913,6 +1092,9 @@ class Window:
     ) -> str:
         return as_text(self.screen, as_ansi, add_history, add_wrap_markers, alternate_screen, add_cursor)
 
+    def cmd_output(self, which: CommandOutput = CommandOutput.last_run, as_ansi: bool = False, add_wrap_markers: bool = False) -> str:
+        return cmd_output(self.screen, which, as_ansi, add_wrap_markers)
+
     @property
     def cwd_of_child(self) -> Optional[str]:
         return self.child.foreground_cwd or self.child.current_cwd
@@ -933,13 +1115,48 @@ class Window:
             'text': text
         }
 
+    def set_logo(self, path: str, position: str = '', alpha: float = -1) -> None:
+        path = resolve_custom_file(path) if path else ''
+        set_window_logo(self.os_window_id, self.tab_id, self.id, path, position or '', alpha)
+
     # actions {{{
 
     @ac('cp', 'Show scrollback in a pager like less')
     def show_scrollback(self) -> None:
         text = self.as_text(as_ansi=True, add_history=True, add_wrap_markers=True)
         data = self.pipe_data(text, has_wrap_markers=True)
-        get_boss().display_scrollback(self, data['text'], data['input_line_number'])
+        cursor_on_screen = self.screen.scrolled_by < self.screen.lines - self.screen.cursor.y
+        get_boss().display_scrollback(self, data['text'], data['input_line_number'], report_cursor=cursor_on_screen)
+
+    def show_cmd_output(self, which: CommandOutput, title: str = 'Command output', as_ansi: bool = True, add_wrap_markers: bool = True) -> None:
+        text = self.cmd_output(which, as_ansi=as_ansi, add_wrap_markers=add_wrap_markers)
+        text = text.replace('\r\n', '\n').replace('\r', '\n')
+        get_boss().display_scrollback(self, text, title=title, report_cursor=False)
+
+    @ac('cp', '''
+        Show output from the first shell command on screen in a pager like less
+
+        Requires :ref:`shell_integration` to work
+        ''')
+    def show_first_command_output_on_screen(self) -> None:
+        self.show_cmd_output(CommandOutput.first_on_screen, 'First command output on screen')
+
+    @ac('cp', '''
+        Show output from the last shell command in a pager like less
+
+        Requires :ref:`shell_integration` to work
+        ''')
+    def show_last_command_output(self) -> None:
+        self.show_cmd_output(CommandOutput.last_run, 'Last command output')
+
+    @ac('cp', '''
+        Show the first command output below the last scrolled position via scroll_to_prompt
+        or the last mouse clicked command output in a pager like less
+
+        Requires :ref:`shell_integration` to work
+        ''')
+    def show_last_visited_command_output(self) -> None:
+        self.show_cmd_output(CommandOutput.last_visited, 'Last visited command output')
 
     def paste_bytes(self, text: Union[str, bytes]) -> None:
         # paste raw bytes without any processing
@@ -970,6 +1187,12 @@ class Window:
         if text:
             set_clipboard_string(text)
 
+    @ac('cp', 'Copy the selected text from the active window to the clipboard with ANSI formatting codes')
+    def copy_ansi_to_clipboard(self) -> None:
+        text = self.text_for_selection(as_ansi=True)
+        if text:
+            set_clipboard_string(text)
+
     def encoded_key(self, key_event: KeyEvent) -> bytes:
         return encode_key_for_tty(
             key=key_event.key, shifted_key=key_event.shifted_key, alternate_key=key_event.alternate_key,
@@ -978,7 +1201,7 @@ class Window:
             cursor_key_mode=self.screen.cursor_key_mode,
         ).encode('ascii')
 
-    @ac('cp', 'Copy the selected text from the active window to the clipboard, if no selection, send Ctrl-C')
+    @ac('cp', 'Copy the selected text from the active window to the clipboard, if no selection, send SIGINT (aka :kbd:`ctrl+c`)')
     def copy_or_interrupt(self) -> None:
         text = self.text_for_selection()
         if text:
@@ -986,7 +1209,7 @@ class Window:
         else:
             self.write_to_child(self.encoded_key(KeyEvent(key=ord('c'), mods=GLFW_MOD_CONTROL)))
 
-    @ac('cp', 'Copy the selected text from the active window to the clipboard and clear selection, if no selection, send Ctrl-C')
+    @ac('cp', 'Copy the selected text from the active window to the clipboard and clear selection, if no selection, send SIGINT (aka :kbd:`ctrl+c`)')
     def copy_and_clear_or_interrupt(self) -> None:
         self.copy_or_interrupt()
         self.screen.clear_selection()
@@ -1001,6 +1224,10 @@ class Window:
             else:
                 open_url(text, cwd=cwd)
 
+    @ac('cp', 'Clear the current selection')
+    def clear_selection(self) -> None:
+        self.screen.clear_selection()
+
     @ac('sc', 'Scroll up by one line')
     def scroll_line_up(self) -> None:
         if self.screen.is_main_linebuf():
@@ -1031,6 +1258,26 @@ class Window:
         if self.screen.is_main_linebuf():
             self.screen.scroll(SCROLL_FULL, False)
 
+    @ac('sc', '''
+        Scroll to the previous/next shell command prompt
+        Allows easy jumping from one command to the next. Requires working
+        :ref:`shell_integration`. Takes a single, optional, number as argument which is
+        the number of prompts to jump, negative values jump up and positive values jump down.
+        A value of zero will jump to the last prompt visited by this action.
+        For example::
+
+            map ctrl+p scroll_to_prompt -1  # jump to previous
+            map ctrl+n scroll_to_prompt 1   # jump to next
+            map ctrl+o scroll_to_prompt 0   # jump to last visited
+        ''')
+    def scroll_to_prompt(self, num_of_prompts: int = -1) -> None:
+        if self.screen.is_main_linebuf():
+            self.screen.scroll_to_prompt(num_of_prompts)
+
+    @ac('sc', 'Scroll prompt to the bottom of the screen, filling in extra lines form the scrollback buffer')
+    def scroll_prompt_to_bottom(self) -> None:
+        self.screen.scroll_prompt_to_bottom()
+
     @ac('mk', 'Toggle the current marker on/off')
     def toggle_marker(self, ftype: str, spec: Union[str, Tuple[Tuple[int, str], ...]], flags: int) -> None:
         from .marks import marker_from_spec
